from .mdd import MDC
from .mdd import MDD
from .marchenko import Marchenko
from .seismicinterpolation import SeismicInterpolation
from .wavedecomposition import PressureToVelocity, UpDownComposition2D, \
    UpDownComposition3D, WavefieldDecomposition
from .oneway import PhaseShift, Deghosting
from .lsm import Demigration, LSM
